import {Component, EventEmitter, Output} from '@angular/core';
import {CommonModule} from '@angular/common';
import {Ingredient} from "../../../Models/ingredient";
import {MatInputModule} from "@angular/material/input";
import {MatAutocompleteModule} from "@angular/material/autocomplete";
import {FormControl, ReactiveFormsModule} from "@angular/forms";
import {map, Observable, startWith} from "rxjs";
import {MatCardModule} from "@angular/material/card";
import {IngredientItemComponent} from "../ingredient-item/ingredient-item.component";
import {MatGridListModule} from "@angular/material/grid-list";
import {MatSelectModule} from "@angular/material/select";
import {FilterService} from "../../../Services/filter.service";
import {IngredientDataService} from "../../../Services/ingredient-data.service";
import {MatProgressSpinnerModule} from "@angular/material/progress-spinner";
import {UnitDataService} from "../../../Services/unit-data.service";
import {Unit} from "../../../Models/unit";

@Component({
  selector: 'app-ingredients-selector',
  standalone: true,
  imports: [CommonModule, MatInputModule, MatAutocompleteModule, ReactiveFormsModule, IngredientItemComponent, MatCardModule, MatGridListModule, MatSelectModule, MatProgressSpinnerModule],
  templateUrl: './ingredients-selector.component.html',
  styleUrl: './ingredients-selector.component.css'
})
export class IngredientsSelectorComponent {

  //TODO make output that has the filter values for vis
  ingredients: Ingredient[] = [];
  units: Unit[] = [];
  ingredientControl = new FormControl();
  filteredOptions: Observable<Ingredient[]> | undefined;
  selectedIngredients: Ingredient[] = [];
  @Output() ingredientChange = new EventEmitter<Ingredient[]>();
  showSpinner: boolean = true;

  constructor(private filterService: FilterService, private ingredientService: IngredientDataService, private unitService: UnitDataService) {
    this.filterService.selectedIngredients$.subscribe(ingredients => {
      this.selectedIngredients = ingredients;
    });
  }

  updateSharedService(): void {

    this.filterService.updateSelectedIngredients(this.selectedIngredients);
  }

  ngOnInit() {
    this.showSpinner = true
    this.ingredientService.getIngredients().subscribe(
      data => {
        this.ingredients = data;

        this.unitService.getUnits().subscribe(
          data => {
            this.units = data
          }
        )

        this.showSpinner = false;
      }
    )
    this.filteredOptions = this.ingredientControl.valueChanges.pipe(
      startWith(''),
      map(value => this._filter(value))
    );
  }

  private _filter(value: string): Ingredient[] {
    const filterValue = (typeof value === 'string' ? value : '').toLowerCase();
    const availableIngredients = this.ingredients.filter(
      ingredient =>
        ingredient.title.toLowerCase().includes(filterValue) &&
        !this.selectedIngredients.some(selected => selected.title === ingredient.title)
    );
    return availableIngredients;
  }

  addIngredient(event: any): void {
    const selectedOption = event.option ? event.option.value : event;
    if (selectedOption) {
      const existingIngredient = this.selectedIngredients.find(
        ingredient => ingredient.title === selectedOption.title
      );
      if (!existingIngredient) {
        this.selectedIngredients.push(selectedOption);
        this.ingredientChange.emit(this.selectedIngredients);
        this.updateSharedService()
      }
      this.ingredientControl.setValue("");
    }
  }
  removeIngredient(event: Ingredient): void {
    const index = this.selectedIngredients.findIndex(i => i === event);

    if (index !== -1) {
      this.selectedIngredients.splice(index, 1);
      this.ingredientChange.emit(this.selectedIngredients);
      this.updateSharedService()
    }
  }

  onAmountChange(amount: number, ingredient: any): void {
    const index = this.selectedIngredients.findIndex(item => item.title === ingredient.title);

    if (index !== -1) {
      this.selectedIngredients[index].amount = amount;
      this.ingredientChange.emit(this.selectedIngredients);
      this.updateSharedService()
    } else {
      console.error('Ingredient not found in the selectedIngredients array:', ingredient);
    }
  }

  onUnitChange(unit: Unit, ingredient: any): void {
    const index = this.selectedIngredients.findIndex(item => item.title === ingredient.title);

    if (index !== -1) {
      this.selectedIngredients[index].selectedUnit = unit;
      this.ingredientChange.emit(this.selectedIngredients);
      this.updateSharedService()
    } else {
      console.error('Ingredient not found in the selectedIngredients array:', ingredient);
    }
  }
}
