import {Component, ElementRef, OnInit, ViewChild} from '@angular/core';
import {CommonModule} from '@angular/common';

import {scatterplot_shader} from '../../../WebGPU/shader/scatterplot_shader';
import {RecipeDataService} from '../../../Services/recipe-data.service'
import { Recipe } from '../../../Models/recipe';
import { fromEvent, max, min, skip } from 'rxjs';
import { VisualisationService } from '../../../Services/visualisation.service';
import { FilterService } from '../../../Services/filter.service';
import { Ingredient } from '../../../Models/ingredient';
import { AxisOptions } from '../../../Models/axis-options';
import { Modes } from '../mode-select/modes';
import { limits_computes_shader } from '../../../WebGPU/shader/limits_compute_shader';

@Component({
  selector: 'app-visualization-area',
  standalone: true,
  imports: [CommonModule],
  templateUrl: './visualization-area.component.html',
  styleUrl: './visualization-area.component.css'
})
export class VisualizationAreaComponent implements OnInit {

  @ViewChild('myCanvas') canvasReference: ElementRef<HTMLCanvasElement> | undefined;
  
  canvas : HTMLCanvasElement | undefined

  device : GPUDevice | undefined
  context : GPUCanvasContext | undefined
  format : GPUTextureFormat | undefined
  size : {width: number, height: number} | undefined

  ratingsBuffer : GPUBuffer | undefined
  complexityBuffer : GPUBuffer | undefined
  statusBuffer : GPUBuffer | undefined
  ingredientsCountBuffer : GPUBuffer | undefined
  ingredientAmountsBuffer : GPUBuffer | undefined
  ingredientUnitsBuffer : GPUBuffer | undefined
  ingredientIDsBuffer : GPUBuffer | undefined
  filterIngredientsBuffer : GPUBuffer | undefined
  indexResultBuffer : GPUBuffer | undefined
  positionsBuffer : GPUBuffer | undefined
  uniformBuffer : GPUBuffer | undefined
  limitsBuffer : GPUBuffer | undefined
  axisBuffer : GPUBuffer | undefined
  limitsClearBuffer : GPUBuffer | undefined

  testBuffer: GPUBuffer | undefined

  ratingLimits : {min: number, max: number} = { min: 0, max: 0 }
  complexityLimits : {min: number, max: number} = { min: 0, max: 0 }
  ingredientCountLimits : {min: number, max: number} = { min: 0, max: 0 }

  bufferSize : number = 0
  indexTexture : GPUTexture | undefined
  indexClearTexture : GPUTexture | undefined
  uniformValues : ArrayBuffer | undefined

  graphicsBindGroup : GPUBindGroup | undefined
  computeBindGroup : GPUBindGroup | undefined
  graphicsPipeline : GPURenderPipeline | undefined
  computePipeline : GPUComputePipeline | undefined

  ngOnInit(): void {}

  constructor(private recipeDataService: RecipeDataService, 
              private visualizationService: VisualisationService, 
              private filterService: FilterService) {}

  
  async ngAfterViewInit() {
    console.log(this.canvasReference)
    if (!this.canvasReference) {
      throw new Error('No Canvas Found')
    }

    this.canvas = this.canvasReference.nativeElement

    if(!navigator.gpu) {
      throw new Error('No Support For WebGPU')
    }

    const adapter = await navigator.gpu.requestAdapter({powerPreference: 'high-performance'})

    if (!adapter) {
      throw new Error('No Adapter Found')
    }

    this.device = await adapter.requestDevice()
    this.context = this.canvas.getContext('webgpu') as GPUCanvasContext
    this.format = navigator.gpu.getPreferredCanvasFormat()
    const devicePixelRatio = window.devicePixelRatio || 1
    this.canvas.width = this.canvas.clientWidth * devicePixelRatio
    this.canvas.height = this.canvas.clientHeight * devicePixelRatio
    this.size = {width: this.canvas.width, height: this.canvas.height}
    this.context.configure({device: this.device, format: this.format, alphaMode: 'opaque'})

    fromEvent(this.canvas, 'click').subscribe((event: Event) => {
      if (event instanceof MouseEvent) {
        this.handleMouseEvent(event); //mousemove
      }
    });

    this.filterService.selectedIngredients$.subscribe(filterIngredients => {
      this.handleFilterChanged(filterIngredients);
    });

    this.visualizationService.selectedAxes$.subscribe(axesSelection => {
      this.handleAxesChanged(axesSelection)
    })

    this.visualizationService.selectedMode$.subscribe(modeSelection => {
      this.handleModeChanged(modeSelection)
    })

    this.recipeDataService.recipes$.pipe(skip(1)).subscribe((result) => {
      this.createBuffers(result);
      this.createTextures();
      this.render();
    });
  }

  async handleMouseEvent(event: MouseEvent) {
    if (!this.canvas || !this.device || !this.indexTexture || !this.size || !this.indexResultBuffer || this.indexResultBuffer.mapState != 'unmapped' ||
        !this.uniformValues || !this.uniformBuffer) {
      return
    }
    const devicePixelRatio = window.devicePixelRatio || 1;
    console.log(this.size);
    var x = Math.floor(event.pageX - this.canvas.getBoundingClientRect().x);
    var y = Math.floor(event.pageY - this.canvas.getBoundingClientRect().y);
    y = this.canvas.getBoundingClientRect().height - y;
    x *= devicePixelRatio;
    y *= devicePixelRatio;
    x = Math.min(this.size.width - 1, Math.max(0, x));
    y = Math.min(this.size.height - 1, Math.max(0, y));

    const encoder = this.device.createCommandEncoder({ label: 'ray cast encoder' });
    encoder.copyTextureToBuffer({texture: this.indexTexture, origin: {x, y}}, {buffer: this.indexResultBuffer, bytesPerRow: 256}, {width: 1, height: 1});
    const commandBuffer = encoder.finish();
    this.device.queue.submit([commandBuffer]);
    await this.indexResultBuffer.mapAsync(GPUMapMode.READ);

    const selectedIndex: number = new Uint32Array(this.indexResultBuffer.getMappedRange())[0];
    this.indexResultBuffer.unmap();

    const selectedIndexView = new Uint32Array(this.uniformValues, 2*4, 1);
    selectedIndexView.set([selectedIndex]);

    this.device.queue.writeBuffer(this.uniformBuffer, 0, this.uniformValues);
    this.draw();

    this.visualizationService.setSelectedRecipe(selectedIndex);
  }

  async handleFilterChanged(filterIngredients: Ingredient[]) {
    if (!this.device || !this.filterIngredientsBuffer) {
      return
    }
    console.log(filterIngredients);
    const filterIngredientBufferData = new Uint32Array(20);
    filterIngredients.forEach((ingredient, ingredientIndex) => {
      filterIngredientBufferData[ingredientIndex] = ingredient['id'];
    });
    console.log(filterIngredientBufferData);
    this.device.queue.writeBuffer(this.filterIngredientsBuffer, 0, filterIngredientBufferData);
    this.draw();
  }

  async handleAxesChanged(axesSelection: { xAxis: AxisOptions; yAxis: AxisOptions, colorAxis: AxisOptions }) {
    if (!this.device || !this.axisBuffer) {
      return
    }
    console.log(axesSelection);

    const axisBufferData = new Uint32Array(3);
    const axisOptions: string[] = Object.values(AxisOptions);

    // 0: random, 1: rating, 2: complexity, 3: ingredient count 
    axisBufferData[0] = axisOptions.indexOf(axesSelection.xAxis);
    axisBufferData[1] = axisOptions.indexOf(axesSelection.yAxis);
    axisBufferData[2] = axisOptions.indexOf(axesSelection.colorAxis);
    
    console.log(axisBufferData);
    this.device.queue.writeBuffer(this.axisBuffer, 0, axisBufferData);
    this.draw();
  }

  async handleModeChanged(modeSelection: Modes) {
    if (!this.device || !this.uniformValues || !this.uniformBuffer) {
      return
    }
    console.log(modeSelection);

    const modeOptions: string[] = Object.values(Modes);

    const selectedModeView = new Uint32Array(this.uniformValues, 3*4, 1);
    selectedModeView.set([modeOptions.indexOf(modeSelection)]);

    this.device.queue.writeBuffer(this.uniformBuffer, 0, this.uniformValues);
    this.draw();
  }

  async createTextures() {
    if (!this.device || !this.size) {
      throw new Error('Missing device for creating buffers!')
    }

    this.indexTexture = this.device.createTexture({
      size: [this.size.width, this.size.height],
      format: 'r32uint',
      usage: GPUTextureUsage.STORAGE_BINDING | GPUTextureUsage.COPY_SRC | GPUTextureUsage.COPY_DST
    })

    this.indexClearTexture = this.device.createTexture({
      size: [this.size.width, this.size.height],
      format: 'r32uint',
      usage: GPUTextureUsage.STORAGE_BINDING | GPUTextureUsage.COPY_SRC
    })

    this.indexResultBuffer = this.device.createBuffer({
      size: 256,
      usage: GPUBufferUsage.COPY_DST | GPUBufferUsage.MAP_READ,
    });
  }

  async createBuffers(recipes: Recipe[]) {
    if (!this.device) {
      throw new Error('Missing device for creating buffers!')
    }

    /* Buffer Layouts:
      # ratingsBuffer:             float32:        rating
      # complexityBuffer:          float32:        complexity
      # statusBuffer:              uint32:         status
      # ingredientCountBuffer:     uint32:         ingredient count
      # ingredientAmountsBuffer:   20 x float32:   amounts
      # ingredientUnitsBuffer:     20 x uint32:    units
      # ingredientIDsBuffer:       20 x 3 x uint32:    ids (3 x 32bit =  6 x 16bit)
    */
    this.bufferSize = recipes.length;
    const ratingsBufferData = new Float32Array(this.bufferSize);
    const complexityBufferData = new Float32Array(this.bufferSize);
    const statusBufferData = new Uint32Array(this.bufferSize);
    const ingredientsCountBufferData = new Uint32Array(this.bufferSize);
    const ingredientAmountsBufferData = new Float32Array(this.bufferSize * 20);
    const ingredientUnitsBufferData = new Uint32Array(this.bufferSize * 20);
    const ingredientIDsBufferData = new Uint32Array(this.bufferSize * 20 * 3);

    const positionsBufferData = new Float32Array(this.bufferSize * 2);

    let unitIdMapping = new Map<string, number>([
        ["undefined", 0],
        ["g", 1],
        ["oz", 2]
    ]);
    
    recipes.forEach((recipe, index) => {
      ratingsBufferData[index] = +recipe['rating'];
      complexityBufferData[index] = recipe['complexity'];
      statusBufferData[index] = recipe['status'];
      ingredientsCountBufferData[index] = recipe['ingredients'].length;

      recipe['ingredients'].forEach((ingredient, ingredientIndex) => {
        ingredientAmountsBufferData[index * 20 + ingredientIndex] = ingredient['amount']
        ingredientUnitsBufferData[index * 20 + ingredientIndex] = unitIdMapping.has(ingredient['unit']) ? unitIdMapping.get(ingredient['unit'])! : 0;
        ingredientIDsBufferData[index * 20 * 3 + ingredientIndex * 3 + 0] = ingredient['ids'][0] << 16 | (ingredient['ids'][1] & 0xff);
        ingredientIDsBufferData[index * 20 * 3 + ingredientIndex * 3 + 1] = ingredient['ids'][2] << 16 | (ingredient['ids'][3] & 0xff);
        ingredientIDsBufferData[index * 20 * 3 + ingredientIndex * 3 + 2] = ingredient['ids'][4] << 16 | (ingredient['ids'][5] & 0xff);
      });

      this.ratingLimits = { min: Math.min(this.ratingLimits.min, ratingsBufferData[index]), 
                            max: Math.max(this.ratingLimits.max, ratingsBufferData[index]) };
      this.complexityLimits = { min: Math.min(this.complexityLimits.min, complexityBufferData[index]), 
                                max: Math.max(this.complexityLimits.max, complexityBufferData[index]) };
      this.ingredientCountLimits = { min: Math.min(this.ingredientCountLimits.min, recipe['ingredients'].length), 
                                     max: Math.max(this.ingredientCountLimits.max, recipe['ingredients'].length) };

      positionsBufferData[index*2 + 0] = Math.random();
      positionsBufferData[index*2 + 1] = Math.random();
    });

    this.ratingsBuffer = this.device.createBuffer({size: this.bufferSize * 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Float32Array(this.ratingsBuffer.getMappedRange()).set(ratingsBufferData);
    this.ratingsBuffer.unmap();
    this.complexityBuffer = this.device.createBuffer({size: this.bufferSize * 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Float32Array(this.complexityBuffer.getMappedRange()).set(complexityBufferData);
    this.complexityBuffer.unmap();
    this.statusBuffer = this.device.createBuffer({size: this.bufferSize * 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Uint32Array(this.statusBuffer.getMappedRange()).set(statusBufferData);
    this.statusBuffer.unmap();
    this.ingredientsCountBuffer = this.device.createBuffer({size: this.bufferSize* 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Uint32Array(this.ingredientsCountBuffer.getMappedRange()).set(ingredientsCountBufferData);
    this.ingredientsCountBuffer.unmap();
    this.ingredientAmountsBuffer = this.device.createBuffer({size: this.bufferSize * 20 * 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Float32Array(this.ingredientAmountsBuffer.getMappedRange()).set(ingredientAmountsBufferData);
    this.ingredientAmountsBuffer.unmap();
    this.ingredientUnitsBuffer = this.device.createBuffer({size: this.bufferSize * 20 * 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Uint32Array(this.ingredientUnitsBuffer.getMappedRange()).set(ingredientUnitsBufferData);
    this.ingredientUnitsBuffer.unmap();
    this.ingredientIDsBuffer = this.device.createBuffer({size: this.bufferSize * 20 * 3 * 4, usage: GPUBufferUsage.STORAGE, mappedAtCreation: true});
    new Uint32Array(this.ingredientIDsBuffer.getMappedRange()).set(ingredientIDsBufferData);
    this.ingredientIDsBuffer.unmap();

    this.filterIngredientsBuffer = this.device.createBuffer({size: 20 * 4, usage: GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_DST});

    this.limitsBuffer = this.device.createBuffer({size: 3 * 4 * 2, usage: GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_SRC | GPUBufferUsage.COPY_DST, mappedAtCreation: true});
    new Uint32Array(this.limitsBuffer.getMappedRange()).set(new Uint32Array([
      this.ratingLimits.min * 1000.0, this.ratingLimits.max * 1000.0,
      this.complexityLimits.min * 1000.0, this.complexityLimits.max * 1000.0,
      this.ingredientCountLimits.min * 1000.0, this.ingredientCountLimits.max * 1000.0,
    ]));
    this.limitsBuffer.unmap();
    this.limitsClearBuffer = this.device.createBuffer({size: 3 * 4 * 2, usage: GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_SRC, mappedAtCreation: true});
    new Uint32Array(this.limitsClearBuffer.getMappedRange()).set(new Uint32Array([
      1000000, 0,
      1000000, 0,
      1000000, 0,
    ]));
    this.limitsClearBuffer.unmap();
    console.log(this.ratingLimits, this.complexityLimits, this.ingredientCountLimits);

    this.axisBuffer = this.device.createBuffer({size: 3 * 4, usage: GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_DST, mappedAtCreation: true});
    // 0: random, 1: rating, 2: complexity, 3: ingredient count 
    new Uint32Array(this.axisBuffer.getMappedRange()).set(new Uint32Array([1,2,0]));
    this.axisBuffer.unmap();

    this.positionsBuffer = this.device.createBuffer({size: this.bufferSize * 4 * 2, usage: GPUBufferUsage.VERTEX | GPUBufferUsage.COPY_DST});
    this.device.queue.writeBuffer(this.positionsBuffer, 0, positionsBufferData);


    this.testBuffer = this.device.createBuffer({
      size: 3*2*4,
      usage: GPUBufferUsage.COPY_DST | GPUBufferUsage.MAP_READ,
    });

  }


  async render() {
    if (!this.device || !this.context || !this.format || !this.size || !this.positionsBuffer || !this.canvas || !this.indexTexture) {
      throw new Error('Initialization Failed!')
    }
    if (!this.ratingsBuffer || !this.ingredientIDsBuffer || !this.filterIngredientsBuffer || 
        !this.limitsBuffer || !this.complexityBuffer || !this.statusBuffer || !this.ingredientsCountBuffer ||
        !this.axisBuffer) {
      throw new Error('Not all buffers were ready to be used as uniforms')
    }

    this.graphicsPipeline = await this.initGraphicsPipeline(this.device, this.format)
    this.computePipeline = await this.initComputePipeline(this.device)

    this.uniformValues = new ArrayBuffer(5 * 4);
    this.uniformBuffer = this.device.createBuffer({
      size: this.uniformValues.byteLength,
      usage: GPUBufferUsage.UNIFORM | GPUBufferUsage.COPY_DST,
    });

    this.graphicsBindGroup = this.device.createBindGroup({
      layout: this.graphicsPipeline.getBindGroupLayout(0),
      entries: [
        { binding: 0, resource: { buffer: this.uniformBuffer } },
        { binding: 1, resource: this.indexTexture.createView() },
        { binding: 2, resource: { buffer: this.ratingsBuffer } },
        { binding: 3, resource: { buffer: this.complexityBuffer} },
        { binding: 4, resource: { buffer: this.statusBuffer} },
        { binding: 5, resource: { buffer: this.ingredientsCountBuffer} },
        { binding: 6, resource: { buffer: this.ingredientIDsBuffer } },
        { binding: 7, resource: { buffer: this.filterIngredientsBuffer } },
        { binding: 8, resource: { buffer: this.limitsBuffer } },
        { binding: 9, resource: { buffer: this.axisBuffer } },
      ],
    });

    this.computeBindGroup = this.device.createBindGroup({
      layout: this.computePipeline.getBindGroupLayout(0),
      entries: [
        { binding: 0, resource: { buffer: this.uniformBuffer } },
        { binding: 1, resource: { buffer: this.ratingsBuffer } },
        { binding: 2, resource: { buffer: this.complexityBuffer} },
        { binding: 3, resource: { buffer: this.ingredientsCountBuffer} },
        { binding: 4, resource: { buffer: this.ingredientIDsBuffer } },
        { binding: 5, resource: { buffer: this.filterIngredientsBuffer } },
        { binding: 6, resource: { buffer: this.limitsBuffer } },
      ],
    });

    const resolutionView = new Float32Array(this.uniformValues, 0, 2);
    resolutionView.set([this.canvas.width, this.canvas.height]);
    const selectedIndexView = new Uint32Array(this.uniformValues, 2*4, 1);
    selectedIndexView.set([3]);
    const selectedModeView = new Uint32Array(this.uniformValues, 3*4, 1);
    selectedModeView.set([0]);
    this.device.queue.writeBuffer(this.uniformBuffer, 0, this.uniformValues);

    this.draw()


    window.addEventListener('resize', ()=>{
      if (!this.canvas) {
        throw new Error('Initialization Failed before Event!')
      }

      this.canvas.width = this.canvas.clientWidth * devicePixelRatio
      this.canvas.height = this.canvas.clientHeight * devicePixelRatio
      // don't need to recall context.configure() after v104
      this.draw()
    })
  }

  
  initGraphicsPipeline(device: GPUDevice, format: GPUTextureFormat): Promise<GPURenderPipeline> {
    const descriptor: GPURenderPipelineDescriptor = {
      layout: 'auto',
      vertex: {
        module: device.createShaderModule({
          code: scatterplot_shader
        }),
        entryPoint: 'vertex',
        buffers: [
          {
            arrayStride: 2 * 4, // 2 floats, 4 bytes each
            stepMode: 'instance',
            attributes: [
              {shaderLocation: 0, offset: 0, format: 'float32x2'},  // position
            ],
          },
        ],
      },
      /*primitive: {
          topology: 'point-list' // try point-list, line-list, line-strip, triangle-strip?
      },*/
      fragment: {
        module: device.createShaderModule({
          code: scatterplot_shader
        }),
        entryPoint: 'fragment',
        targets: [
          {
            format: format,
            blend: {
              color: { operation: "add", srcFactor: "src-alpha", dstFactor: "one-minus-src-alpha" },
              alpha: { operation: "add", srcFactor: "src-alpha", dstFactor: "one-minus-src-alpha" }
            }
          }
        ]
      }
    }
    return device.createRenderPipelineAsync(descriptor)
  }

  initComputePipeline(device: GPUDevice): Promise<GPUComputePipeline> {
    const descriptor: GPUComputePipelineDescriptor = {
      layout: 'auto',
      compute: {
        module: device.createShaderModule({ code: limits_computes_shader }),
        entryPoint: 'compute'
      }
    }
    return device.createComputePipelineAsync(descriptor);
  }


  async draw() {
    if (!this.graphicsBindGroup || !this.computeBindGroup || !this.uniformValues || !this.uniformBuffer || !this.device || !this.graphicsPipeline ||
        !this.computePipeline || !this.context || !this.positionsBuffer || !this.limitsBuffer || !this.limitsClearBuffer || !this.indexTexture ||
        !this.indexClearTexture || !this.size) {
      throw new Error('Not everything ready for draw');
    }

    //this.device.queue.writeBuffer(this.limitsBuffer, 0, new Uint32Array(3*2));

    const commandEncoder = this.device.createCommandEncoder()
    const view = this.context.getCurrentTexture().createView()
    const renderPassDescriptor: GPURenderPassDescriptor = {
      colorAttachments: [
        {
          view: view,
          clearValue: { r: 1, g: 1, b: 1, a: 1 },
          loadOp: 'clear', // clear/load
          storeOp: 'store' // store/discard
        }
      ]
    }

    commandEncoder.copyTextureToTexture({texture: this.indexClearTexture}, {texture: this.indexTexture}, [this.size.width, this.size.height]);
    commandEncoder.copyBufferToBuffer(this.limitsClearBuffer, 0, this.limitsBuffer, 0, 3*2*4);

    const computePass = commandEncoder.beginComputePass();
    computePass.setPipeline(this.computePipeline);
    computePass.setBindGroup(0, this.computeBindGroup);
    computePass.dispatchWorkgroups(this.bufferSize / 64);
    computePass.end();




    // if (!this.testBuffer) {
    //   throw new Error('Bad!');
    // }
    // const encoder = this.device.createCommandEncoder({ label: 'some encoder' });
    // encoder.copyBufferToBuffer(this.limitsBuffer, 0, this.testBuffer, 0, 3*2*4);
    // const commandBuffer = encoder.finish();
    // this.device.queue.submit([commandBuffer]);

    
    // await this.testBuffer.mapAsync(GPUMapMode.READ);
    // const values = new Uint32Array(this.testBuffer.getMappedRange());
    // console.log(values);
    // this.testBuffer.unmap();


    const passEncoder = commandEncoder.beginRenderPass(renderPassDescriptor)
    passEncoder.setPipeline(this.graphicsPipeline)
    passEncoder.setVertexBuffer(0, this.positionsBuffer);
    passEncoder.setBindGroup(0, this.graphicsBindGroup);
    passEncoder.draw(6, this.bufferSize);
    passEncoder.end()
    // webgpu run in a separate process, all the commands will be executed after submit
    this.device.queue.submit([commandEncoder.finish()])
  }
}
