import {Component, ElementRef, OnInit, ViewChild} from '@angular/core';
import {CommonModule} from '@angular/common';

import {scatterplot_shader} from '../../WebGPU/shader/scatterplot_shader';


@Component({
  selector: 'app-web-gpu-test',
  standalone: true,
  imports: [CommonModule],
  templateUrl: './web-gpu-test.component.html',
  styleUrl: './web-gpu-test.component.css'
})
export class WebGpuTestComponent implements OnInit {

  @ViewChild('myCanvas') canvasReference: ElementRef<HTMLCanvasElement> | undefined;
  
  canvas : HTMLCanvasElement | undefined

  device : GPUDevice | undefined
  context : GPUCanvasContext | undefined
  format : GPUTextureFormat | undefined
  size : {width: number, height: number} | undefined


  ngOnInit(): void {}

  
  async ngAfterViewInit() {
    console.log(this.canvasReference)
    if (!this.canvasReference) {
      throw new Error('No Canvas Found')
    }

    this.canvas = this.canvasReference.nativeElement

    if(!navigator.gpu) {
      throw new Error('No Support For WebGPU')
    }

    const adapter = await navigator.gpu.requestAdapter({powerPreference: 'high-performance'})

    if (!adapter) {
      throw new Error('No Adapter Found')
    }

    this.device = await adapter.requestDevice()
    this.context = this.canvas.getContext('webgpu') as GPUCanvasContext
    this.format = navigator.gpu.getPreferredCanvasFormat()
    const devicePixelRatio = window.devicePixelRatio || 1
    this.canvas.width = this.canvas.clientWidth * devicePixelRatio
    this.canvas.height = this.canvas.clientHeight * devicePixelRatio
    this.size = {width: this.canvas.width, height: this.canvas.height}
    this.context.configure({device: this.device, format: this.format, alphaMode: 'opaque'})

    this.render()
  }


  async render() {
    if (!this.device || !this.context || !this.format || !this.size) {
        throw new Error('Initialization Failed!')
    }


    const pipeline = await this.initPipeline(this.device, this.format)

    console.log(this.device, this.context, pipeline)
    this.draw(this.device, this.context, pipeline)
    
    window.addEventListener('resize', ()=>{
      if (!this.device || !this.context || !this.canvas) {
        throw new Error('Initialization Failed before Event!')
      }

      this.canvas.width = this.canvas.clientWidth * devicePixelRatio
      this.canvas.height = this.canvas.clientHeight * devicePixelRatio
      // don't need to recall context.configure() after v104
      this.draw(this.device, this.context, pipeline)
    })
  }

  
  initPipeline(device: GPUDevice, format: GPUTextureFormat): Promise<GPURenderPipeline> {
      const descriptor: GPURenderPipelineDescriptor = {
          layout: 'auto',
          vertex: {
              module: device.createShaderModule({
                  code: scatterplot_shader
              }),
              entryPoint: 'vertex'
          },
          primitive: {
              topology: 'triangle-list' // try point-list, line-list, line-strip, triangle-strip?
          },
          fragment: {
              module: device.createShaderModule({
                  code: scatterplot_shader
              }),
              entryPoint: 'fragment',
              targets: [
                  {
                      format: format
                  }
              ]
          }
      }
      return device.createRenderPipelineAsync(descriptor)
  }


  draw(device: GPUDevice, context: GPUCanvasContext, pipeline: GPURenderPipeline) {
      const commandEncoder = device.createCommandEncoder()
      const view = context.getCurrentTexture().createView()
      const renderPassDescriptor: GPURenderPassDescriptor = {
          colorAttachments: [
              {
                  view: view,
                  clearValue: { r: 0, g: 0, b: 0, a: 1.0 },
                  loadOp: 'clear', // clear/load
                  storeOp: 'store' // store/discard
              }
          ]
      }
      const passEncoder = commandEncoder.beginRenderPass(renderPassDescriptor)
      passEncoder.setPipeline(pipeline)
      // 3 vertex form a triangle
      passEncoder.draw(3)
      passEncoder.end()
      // webgpu run in a separate process, all the commands will be executed after submit
      device.queue.submit([commandEncoder.finish()])
  }
}
