export const scatterplot_shader = /* wgsl */ `

struct Vertex {
  @location(0) position: vec2f,
};

struct Uniforms {
  resolution: vec2f,
  selected_index: u32,
  selected_mode: u32,
};

struct IntLimits {
  rating_min: u32,
  rating_max: u32,
  complexity_min: u32,
  complexity_max: u32,
  ingredient_count_min: u32,
  ingredient_count_max: u32,
}

struct Limits {
  rating_min: f32,
  rating_max: f32,
  complexity_min: f32,
  complexity_max: f32,
  ingredient_count_min: f32,
  ingredient_count_max: f32,
}

struct VSOutput {
  @builtin(position) position: vec4f,
  @location(0) texcoord: vec2f,
  @location(1) @interpolate(flat) index: u32,
  @location(2) screen_uv: vec2f,
  @location(3) @interpolate(flat) visible: u32,
  @location(4) @interpolate(flat) color_axis_value: f32,
};

@group(0) @binding(0) var<uniform> uniforms: Uniforms;
@group(0) @binding(1) var indices: texture_storage_2d<r32uint, write>;
@group(0) @binding(2) var<storage, read> rating: array<f32>;
@group(0) @binding(3) var<storage, read> complexity: array<f32>;
@group(0) @binding(4) var<storage, read> status: array<u32>;
@group(0) @binding(5) var<storage, read> ingredients_count: array<u32>;
@group(0) @binding(6) var<storage, read> ingredient_ids: array<u32>;
@group(0) @binding(7) var<storage, read> filter_ingredient_ids: array<u32>;
@group(0) @binding(8) var<storage, read> int_limits: IntLimits;
@group(0) @binding(9) var<storage, read> axes: array<u32>;

const pointScale = 0.02;


fn test_recipe(id: u32, explore_mode: bool) -> bool {
  var found_ingredients = array<bool, 20>();

  for (var i:u32=0u; i<20u; i++ ) {
    var ingredient_part_01:u32 = ingredient_ids[id*20*3 + i*3 + 0];
    var ingredient_part_23:u32 = ingredient_ids[id*20*3 + i*3 + 1];
    var ingredient_part_45:u32 = ingredient_ids[id*20*3 + i*3 + 2];

    var ingredient_0:u32 = ingredient_part_01 >> 16u;
    var ingredient_1:u32 = ingredient_part_01 & 0xff;
    var ingredient_2:u32 = ingredient_part_23 >> 16u;
    var ingredient_3:u32 = ingredient_part_23 & 0xff;
    var ingredient_4:u32 = ingredient_part_45 >> 16u;
    var ingredient_5:u32 = ingredient_part_45 & 0xff;

    for (var j:u32=0u; j<20u; j++ ) {
      var filter_ingredient = filter_ingredient_ids[j];

      if (filter_ingredient == 0) {
        continue;
      }

      if (ingredient_0 == filter_ingredient || 
          ingredient_1 == filter_ingredient || 
          ingredient_2 == filter_ingredient || 
          ingredient_3 == filter_ingredient || 
          ingredient_4 == filter_ingredient || 
          ingredient_5 == filter_ingredient) {
            
          found_ingredients[j] = true;
      }
    }
  }


  for (var i:u32=0u; i<20u; i++ ) {
    if (filter_ingredient_ids[i] == 0) {
      continue;
    }
    if (explore_mode && found_ingredients[i]) {
      return true;
    } else if(!explore_mode && !found_ingredients[i]) {
      return false;
    }
  }

  return !explore_mode;
}


@vertex
fn vertex(
  @builtin(vertex_index) vertex_index: u32,
  @builtin(instance_index) instance_index: u32,
  vert: Vertex,
) -> VSOutput {
  // quad
  let points = array(
    vec2f(-1, -1),
    vec2f( 1, -1),
    vec2f(-1,  1),
    vec2f(-1,  1),
    vec2f( 1, -1),
    vec2f( 1,  1),
  );

  var instance_position = vert.position;

  var instance_rating = rating[instance_index];
  var instance_complexity = complexity[instance_index];
  var instance_status = status[instance_index];
  var instance_ingredients_count = ingredients_count[instance_index];

  var limits = Limits(
    0.001 * f32(int_limits.rating_min),
    0.001 * f32(int_limits.rating_max),
    0.001 * f32(int_limits.complexity_min),
    0.001 * f32(int_limits.complexity_max),
    0.001 * f32(int_limits.ingredient_count_min),
    0.001 * f32(int_limits.ingredient_count_max),
  );

  var normalized_rating = (instance_rating - limits.rating_min) / limits.rating_max;
  var normalized_complexity = (instance_complexity - limits.complexity_min) / limits.complexity_max;
  var normalized_ingredients_count = (f32(instance_ingredients_count) - limits.ingredient_count_min) / limits.ingredient_count_max;

  var axis_values = array<f32, 4>(
    0,
    normalized_rating,
    normalized_complexity,
    normalized_ingredients_count,
  );

  if (axes[0] != 0) {
    instance_position.x = axis_values[axes[0]];
  }
  if (axes[1] != 0) {
    instance_position.y = axis_values[axes[1]];
  }

  var aspectRatio = uniforms.resolution.x / uniforms.resolution.y;
  var uv = instance_position * 2.0 - 1.0;
  
  var position = uv + points[vertex_index] * pointScale;
  position.y *= aspectRatio;

  position *= 1.0 - pointScale;

  
  var vsOut: VSOutput;
  vsOut.position = vec4f(position, 0.0, 1.0);
  vsOut.texcoord = points[vertex_index] * 0.5 + 0.5;
  vsOut.index = instance_index;
  vsOut.screen_uv = position * 0.5 + 0.5;
  vsOut.visible = u32(test_recipe(instance_index, uniforms.selected_mode == 0));
  vsOut.color_axis_value = 1.0;
  if (axes[2] != 0) {
    vsOut.color_axis_value = axis_values[axes[2]];
  }

  return vsOut;
}

@fragment
fn fragment(vsOut: VSOutput) -> @location(0) vec4f {
  if (vsOut.visible == 0) {
    return vec4(0,0,0,0);
  }

  var dist = length(vsOut.texcoord*2.0-1.0);
  var ratingValue = rating[vsOut.index];

  var screen_uv = vsOut.screen_uv;
  
  if (dist < 0.9) {
    textureStore(indices, vec2u(screen_uv * uniforms.resolution), vec4(vsOut.index));
  }

  var fillClr = mix(vec4(255,255,255,180)/255.0, vec4(41,182,255,180)/255.0, vsOut.color_axis_value);
  var clr = mix(vec4(0,144,219,255)/255.0, fillClr, smoothstep(1.0, 0.6, dist));
  if (vsOut.index == uniforms.selected_index && uniforms.selected_index != 0) {
    clr = vec4(254,205,26,255)/255.0;
  }

  return vec4(clr.rgb, clr.a * smoothstep(1.0, 0.9, dist));
}

`