import ast
import csv
import re
from pathlib import Path
import json
from tqdm import tqdm

fractions = {'½': 1/2, '¼': 1/4, '⅓': 1/3, '⅔': 2/3, '¾': 3/4, '⅛': 1/8}
multi_char_fractions = {'1\u20442' : '½', '1\u20444' : '¼', '1\u20443' : '⅓', '2\u20443' : '⅔', '3\u20444' : '¾', '1\u20448' : '⅛'}
replacements = {'ﬂ': 'fl', 'ﬀ': 'ff', 'ﬁ': 'fi', '\"': 'in', '\u201d': 'in'}
numbers = {'one' : 1, 'two' : 2, 'three' : 3, 'four' : 4, 'five' : 5, 'six' : 6, 'seven' : 7, 'eight' : 8, 'nine' : 9, 'ten' : 10, 'eleven' : 11, 'twelve' : 12}

with open(Path(__file__).parent / 'results' / 'units.json', 'r') as json_file:
    units = json.load(json_file)

with open(Path(__file__).parent / 'results' / 'ingredients.json', 'r') as json_file:
    ingredients_index = json.load(json_file)

class Ingredient:
    def __init__(self, name: str, indices: list, amount: int|float, unit: str, flag: bool) -> None:
        self.name = name
        self.indices = indices
        self.amount = amount
        self.unit = unit
        self.flag = flag
        self.vegan = all(ingredients_index[index]['status'] == 'vegan' for index in self.indices)
        self.vegetarian = all(ingredients_index[index]['status'] == 'vegetarian' for index in self.indices)
        self.meat = all(ingredients_index[index]['status'] == 'meat' for index in self.indices)

    def as_dict(self) -> dict:
        return {
            'ids' : [int(index) for index in self.indices],
            'amount' : self.amount,
            'unit' : self.unit
        }


class IngredientFactory:
    ingredients_index: dict
    lookup : dict
    missing: dict

    def is_amount(word: str) -> bool:
        return re.match('\d', word) or \
               word in numbers or \
               any([fraction in word for fraction in fractions.keys()])
    
    def is_unit(word: str) -> bool:
        return word in units

    
    def parse_amount(word: str, current_amount: int|float, current_unit: str, line) -> tuple[int|float, str]:
        # don't override the first amount and unit seen
        if current_amount != 'unknown' or current_unit != 'unknown':
            return current_amount, current_unit

        if '/' in word:
            word = word.split('/')[0]
        
        for multi_char_fraction, single_char_fraction  in multi_char_fractions.items():
            word = word.replace(multi_char_fraction, single_char_fraction)


        # match e.g. 'one', 'three', ...
        if word in numbers:
            return (numbers[word], 'pcs') # pieces
        
        # match e.g. '1', '3.2', 5g, 2¼oz, ...
        match = re.match('^([0-9.,]+)?([½¼⅓⅔¾⅛])?(\w+)?$', word) 
        if match:
            amount = 0
            unit = 'unknown'
            if match.group(1):
                try:
                    try:
                        amount += int(match.group(1))
                    except ValueError:
                        amount += float(match.group(1))
                except ValueError:
                    amount = 'unknown'
            if match.group(2):
                amount += fractions[match.group(2)]
            if match.group(3):
                unit = match.group(3)
            return (amount, unit)
        
        # match e.g. '2-3', '8-10tbsp', '1g-5g', ...
        match = re.match('^([0-9.,]+)?([½¼⅓⅔¾⅛])?(\w+)?[-\u2013]([0-9.,]+)?([½¼⅓⅔¾⅛])?(\w+)?$', word) 
        if match:
            start = 0
            end = 0
            start_unit = 'unknown'
            end_unit = 'unknown'
            try:
                if match.group(1):
                    start += float(match.group(1))
                if match.group(2):
                    start += fractions[match.group(2)]
                if match.group(3):
                    start_unit = units[match.group(3)]
                if match.group(4):
                    end += float(match.group(4))
                if match.group(5):
                    end += fractions[match.group(5)]
                if match.group(6):
                    end_unit = units[match.group(6)]
            except KeyError:
                pass

            if start_unit != 'unknown' and end_unit != 'unknown' and start_unit == end_unit: # e.g. '5g-8g'
                return ((start + end) / 2, start_unit)
            if start_unit != 'unknown' and end_unit != 'unknown' and start_unit != end_unit: # e.g. '800g-1kg'
                return (start, start_unit)
            if start_unit == 'unknown' and end_unit != 'unknown' and start != 0 and end != 0: # e.g. '3-5tbsp'
                return ((start + end) / 2, end_unit)
            if start_unit == 'unknown' and end_unit != 'unknown' and start != 0 and end == 0: # e.g. '3-litres'
                return (start, end_unit)
            if start_unit == 'unknown' and end_unit == 'unknown' and start != 0 and end != 0: # e.g. '4-5'
                return (start, end_unit)

            return ('unknown', 'unknown')
    
        #print(f'\"{word}\" \t-> {line}')
        return ('unknown', 'unknown')
        
    def parse_unit(word: str) -> str:
        return units[word]


    @classmethod
    def initialize(cls) -> None:
        cls.lookup = {ingredient['name'] : index for index, ingredient in ingredients_index.items()}
        cls.uses = {}
        cls.missing = {}

    @classmethod
    def construct_ingredient(cls, name) -> Ingredient:
        name = name.strip()
        name = name.translate(str.maketrans('', '', '(),+')) # remove unnecessary characters
        name = name.lower() # e.g. '350g\/12oz penne pasta'
        words = name.split(' ') # e.g. ['350g\/12oz', 'penne', 'pasta']

        # results:
        matching_indices = []
        unit = 'unknown'
        amount = 'unknown'

        for word in words:
            word = word.strip()
            word = word.translate(str.maketrans('', '', '\u200b'))
            word = ''.join([replacements[char] if char in replacements else char for char in word])

            if cls.is_amount(word): # either a number or a number with a unit
                amount, unit = cls.parse_amount(word, amount, unit, name)
            elif cls.is_unit(word) and unit == 'unknown': # a simple unit
                unit = cls.parse_unit(word)
            elif word in cls.lookup: # anything else will by compared with the lookup table of ingredient names
                matching_indices.append(cls.lookup[word])
        
        if not matching_indices:
            for word in words:
                if cls.is_amount(word) or cls.is_unit(word):
                    continue
                if word in cls.missing:
                    cls.missing[word] += 1
                else:
                    cls.missing[word] = 1
            return None
        
        return Ingredient(name, matching_indices, amount, unit, False)



class Recipe:
    def __init__(self, line: list) -> None:
        self.id = int(line[0])
        self.title = line[1]
        self.title_lower = line[2]
        self.ingredients = [IngredientFactory.construct_ingredient(i.strip()) for i in ast.literal_eval(line[3]) if i.strip()]
        self.ingredients = [ingredient for ingredient in self.ingredients if ingredient]
        self.directions = line[4]
        self.complexity = len(self.directions)
        self.rating = float(line[5]) if line[5] else None
        self.vegan = all([ingredient.vegan for ingredient in self.ingredients])
        self.vegetarian = all([ingredient.vegetarian for ingredient in self.ingredients])
        self.meat = all([ingredient.meat for ingredient in self.ingredients])

    def __str__(self) -> str:
        return f'{self.id}'
    
    def as_dict(self) -> dict:
        return {
            'title' : self.title,
            'title_lower' : self.title_lower,
            'ingredients' : [ingredient.as_dict() for ingredient in self.ingredients],
            'rating' : self.rating,
            'complexity' : self.complexity,
            'status' : int(''.join(map(str, map(int, [self.vegan, self.vegetarian, self.meat]))), 2) # e.g. [False True False] -> [0 1 0] -> 2
        }



def read_recipes(filename: Path):
    print(filename)
    with open(filename, 'r') as fp:
        reader = csv.reader(fp)
        next(reader) # skip header
        all_recipes_dict = dict()
        for line in tqdm(reader):
            recipe = Recipe(line)
            all_recipes_dict[recipe.id] = recipe.as_dict()

    print('number of recipes:', len(all_recipes_dict))
    return all_recipes_dict



IngredientFactory.initialize()

filename0 = Path(__file__).parent / 'data' / 'new_crawl_recipe(basic).csv'
filename1 = Path(__file__).parent / 'data' / 'original_data(basic).csv'

all_recipes_dict = read_recipes(filename0)

with open('./results/recipes.json', 'w') as json_file:
    json.dump(all_recipes_dict, json_file)

with open('./tmp/missing_ingredients.json', 'w') as json_file:
    json.dump(IngredientFactory.missing, json_file, indent=4, sort_keys=True)


